@@ -46,7 +46,8 @@ struct usb_dev_handle {
 };
 
 /* descriptors.c */
-int usb_parse_configuration(struct usb_config_descriptor *config, char *buffer);
+int usb_parse_configuration(struct usb_config_descriptor *config,
+	unsigned char *buffer);
 void usb_fetch_and_parse_descriptors(usb_dev_handle *udev);
 void usb_destroy_configuration(struct usb_device *dev);
 
